@@ -3,10 +3,13 @@
 import datetime
 from dateutil.relativedelta import relativedelta
 import itertools
+from sql.aggregate import Sum
+from sql.conditionals import Coalesce
+
 from trytond.model import ModelView, Workflow, ModelSQL, fields
 from trytond.wizard import Wizard, StateView, StateTransition, Button
 from trytond.pyson import Not, Equal, Eval, Or, Bool, If
-from trytond.backend import TableHandler
+from trytond import backend
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 from trytond.tools import reduce_ids
@@ -28,9 +31,9 @@ class Forecast(Workflow, ModelSQL, ModelView):
         'stock.location', 'Location', required=True,
         domain=[('type', '=', 'warehouse')], states={
             'readonly': Or(Not(Equal(Eval('state'), 'draft')),
-                Bool(Eval('lines'))),
+                Bool(Eval('lines', [0]))),
             },
-        depends=['state', 'lines'])
+        depends=['state'])
     destination = fields.Many2One(
         'stock.location', 'Destination', required=True,
         domain=[('type', 'in', ['customer', 'production'])], states=STATES,
@@ -45,9 +48,9 @@ class Forecast(Workflow, ModelSQL, ModelView):
     company = fields.Many2One(
         'company.company', 'Company', required=True, states={
             'readonly': Or(Not(Equal(Eval('state'), 'draft')),
-                Bool(Eval('lines'))),
+                Bool(Eval('lines', [0]))),
             },
-        depends=['state', 'lines'])
+        depends=['state'])
     state = fields.Selection([
         ('draft', 'Draft'),
         ('done', 'Done'),
@@ -94,7 +97,9 @@ class Forecast(Workflow, ModelSQL, ModelView):
     @classmethod
     def __register__(cls, module_name):
         Location = Pool().get('stock.location')
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
 
         table = TableHandler(cursor, cls, module_name)
         migrate_warehouse = (not table.column_exist('warehouse')
@@ -114,7 +119,7 @@ class Forecast(Workflow, ModelSQL, ModelView):
                     return location.id
                 elif location.parent:
                     return find_warehouse(location.parent)
-            cursor.execute('SELECT id, location FROM "%s"' % cls._table)
+            cursor.execute(*sql_table.select(sql_table.id, sql_table.location))
             for forecast_id, location_id in cursor.fetchall():
                 warehouse_id = location_id  # default fallback
                 if location_id in location2warehouse:
@@ -123,9 +128,10 @@ class Forecast(Workflow, ModelSQL, ModelView):
                     location = Location(location_id)
                     warehouse_id = find_warehouse(location) or location_id
                     location2warehouse[location_id] = warehouse_id
-                cursor.execute('UPDATE "%s" SET warehouse = %%s '
-                    'WHERE id = %%s' % cls._table,
-                    (warehouse_id, forecast_id))
+                cursor.execute(*sql_table.update(
+                        columns=[sql_table.warehouse],
+                        values=[warehouse_id],
+                        where=sql_table.id == forecast_id))
             table.not_null_action('warehouse',
                 action=cls.warehouse.required and 'add' or 'remove')
             table.drop_column('location', True)
@@ -158,22 +164,19 @@ class Forecast(Workflow, ModelSQL, ModelView):
     def check_date_overlap(self):
         cursor = Transaction().cursor
         if self.state != 'done':
-            return True
-        cursor.execute('SELECT id '
-            'FROM stock_forecast '
-            'WHERE ((from_date <= %s AND to_date >= %s) '
-                    'OR (from_date <= %s AND to_date >= %s) '
-                    'OR (from_date >= %s AND to_date <= %s)) '
-                'AND warehouse = %s '
-                'AND destination = %s '
-                'AND state = \'done\' '
-                'AND company = %s '
-                'AND id != %s',
-            (self.from_date, self.from_date,
-             self.to_date, self.to_date,
-             self.from_date, self.to_date,
-             self.warehouse.id, self.destination.id,
-             self.company.id, self.id))
+            return
+        forcast = self.__table__()
+        cursor.execute(*forcast.select(forcast.id,
+                where=(((forcast.from_date <= self.from_date)
+                        & (forcast.to_date >= self.from_date))
+                    | ((forcast.from_date <= self.to_date)
+                        & (forcast.to_date >= self.to_date))
+                    | ((forcast.from_date >= self.from_date)
+                        & (forcast.to_date <= self.to_date)))
+                & (forcast.warehouse == self.warehouse.id)
+                & (forcast.destination == self.destination.id)
+                & (forcast.company == self.company.id)
+                & (forcast.id != self.id)))
         forecast_id = cursor.fetchone()
         if forecast_id:
             second = self.__class__(forecast_id[0])
@@ -257,18 +260,16 @@ class ForecastLine(ModelSQL, ModelView):
         domain=[
             ('type', '=', 'goods'),
             ('consumable', '=', False),
-            ],
-        on_change=['product'])
+            ])
     product_uom_category = fields.Function(
-        fields.Many2One('product.uom.category', 'Product Uom Category',
-            on_change_with=['product']),
+        fields.Many2One('product.uom.category', 'Product Uom Category'),
         'on_change_with_product_uom_category')
     uom = fields.Many2One('product.uom', 'UOM', required=True,
         domain=[
             If(Bool(Eval('product_uom_category')),
                 ('category', '=', Eval('product_uom_category')),
                 ('category', '!=', -1)),
-            ], on_change=['uom'], depends=['product', 'product_uom_category'])
+            ], depends=['product', 'product_uom_category'])
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
             'get_unit_digits')
     quantity = fields.Float('Quantity', digits=(16, Eval('unit_digits', 2)),
@@ -305,6 +306,7 @@ class ForecastLine(ModelSQL, ModelView):
     def default_minimal_quantity():
         return 1.0
 
+    @fields.depends('product')
     def on_change_product(self):
         res = {}
         res['unit_digits'] = 2
@@ -314,10 +316,12 @@ class ForecastLine(ModelSQL, ModelView):
             res['unit_digits'] = self.product.default_uom.digits
         return res
 
+    @fields.depends('product')
     def on_change_with_product_uom_category(self, name=None):
         if self.product:
             return self.product.default_uom_category.id
 
+    @fields.depends('uom')
     def on_change_uom(self):
         res = {}
         res['unit_digits'] = 2
@@ -338,6 +342,11 @@ class ForecastLine(ModelSQL, ModelView):
         Forecast = pool.get('stock.forecast')
         LineMove = pool.get('stock.forecast.line-stock.move')
 
+        move = Move.__table__()
+        location_from = Location.__table__()
+        location_to = Location.__table__()
+        line_move = LineMove.__table__()
+
         result = dict((x.id, 0) for x in lines)
         key = lambda line: line.forecast.id
         lines.sort(key=key)
@@ -347,28 +356,26 @@ class ForecastLine(ModelSQL, ModelView):
             product_ids = product2line.keys()
             for i in range(0, len(product_ids), cursor.IN_MAX):
                 sub_ids = product_ids[i:i + cursor.IN_MAX]
-                red_sql, red_ids = reduce_ids('product', sub_ids)
-                cursor.execute('SELECT m.product, '
-                        'SUM(m.internal_quantity) AS quantity '
-                    'FROM "' + Move._table + '" AS m '
-                        'JOIN "' + Location._table + '" AS fl '
-                            'ON m.from_location = fl.id '
-                        'JOIN "' + Location._table + '" AS tl '
-                            'ON m.to_location = tl.id '
-                        'LEFT JOIN "' + LineMove._table + '" AS lm '
-                            'ON m.id = lm.move '
-                    'WHERE ' + red_sql + ' '
-                        'AND fl.left >= %s AND fl.right <= %s '
-                        'AND tl.left >= %s AND tl.right <= %s '
-                        'AND m.state != %s '
-                        'AND COALESCE(m.effective_date, m.planned_date) >= %s '
-                        'AND COALESCE(m.effective_date, m.planned_date) <= %s '
-                        'AND lm.id IS NULL '
-                    'GROUP BY m.product',
-                    red_ids + [forecast.warehouse.left,
-                        forecast.warehouse.right, forecast.destination.left,
-                        forecast.destination.right, 'cancel',
-                        forecast.from_date, forecast.to_date])
+                red_sql = reduce_ids(move.product, sub_ids)
+                cursor.execute(*move.join(location_from,
+                        condition=move.from_location == location_from.id
+                        ).join(location_to,
+                        condition=move.to_location == location_to.id
+                        ).join(line_move, 'LEFT',
+                        condition=move.id == line_move.move
+                        ).select(move.product, Sum(move.internal_quantity),
+                        where=red_sql
+                        & (location_from.left >= forecast.warehouse.left)
+                        & (location_from.right <= forecast.warehouse.right)
+                        & (location_to.left >= forecast.destination.left)
+                        & (location_to.right <= forecast.destination.right)
+                        & (move.state != 'cancel')
+                        & (Coalesce(move.effective_date, move.planned_date)
+                            >= forecast.from_date)
+                        & (Coalesce(move.effective_date, move.planned_date)
+                            <= forecast.to_date)
+                        & (line_move.id == None),
+                        group_by=move.product))
                 for product_id, quantity in cursor.fetchall():
                     line = product2line[product_id]
                     result[line.id] = Uom.compute_qty(line.product.default_uom,
@@ -431,7 +438,7 @@ class ForecastLine(ModelSQL, ModelView):
         moves = []
         if to_create:
             moves = Move.create(to_create)
-        self.write([self], {'moves': [('set', [m.id for m in moves])]})
+        self.write([self], {'moves': [('add', [m.id for m in moves])]})
 
     @classmethod
     def delete_moves(cls, lines):
@@ -513,7 +520,8 @@ class ForecastComplete(Wizard):
     def __setup__(cls):
         super(ForecastComplete, cls).__setup__()
         cls._error_messages.update({
-                'from_to_date': '"From Date" should be smaller than "To Date".',
+                'from_to_date': (
+                    '"From Date" should be smaller than "To Date".'),
                 })
 
     def default_ask(self, fields):
@@ -537,11 +545,11 @@ class ForecastComplete(Wizard):
             self.raise_user_error('from_to_date')
 
         with Transaction().set_context(
-                stock_destination=[forecast.destination.id],
+                stock_destinations=[forecast.destination.id],
                 stock_date_start=self.ask.from_date,
                 stock_date_end=self.ask.to_date):
             return Product.products_by_location([forecast.warehouse.id],
-                    with_childs=True, skip_zero=False)
+                with_childs=True)
 
     def default_choose(self, fields):
         """
